@@ -32,7 +32,6 @@
 #include <sys/types.h>
 #include <pwd.h>
 
-#include <polkit/polkit.h>
 #include <systemd/sd-login.h>
 #include <systemd/sd-daemon.h>
 
@@ -40,6 +39,7 @@
 #include <glib-object.h>
 #include <glib/gi18n.h>
 #include <gio/gio.h>
+#include <gio/gunixfdlist.h>
 
 #include "gsm-marshal.h"
 #include "gsm-system.h"
@@ -55,8 +55,9 @@ struct _GsmSystemdPrivate
         GDBusProxy      *sd_proxy;
         gchar           *session_id;
         gchar           *session_path;
-        PolkitAuthority *authority;
-        PolkitSubject   *subject;
+
+        GSList          *inhibitors;
+        gint             inhibit_fd;
 };
 
 static void gsm_systemd_system_init (GsmSystemInterface *iface);
@@ -72,11 +73,14 @@ gsm_systemd_finalize (GObject *object)
         GsmSystemd *systemd = GSM_SYSTEMD (object);
 
         g_clear_object (&systemd->priv->sd_proxy);
-        g_clear_object (&systemd->priv->authority);
-        g_clear_object (&systemd->priv->subject);
         g_free (systemd->priv->session_id);
         g_free (systemd->priv->session_path);
 
+        if (systemd->priv->inhibitors != NULL) {
+                g_slist_free_full (systemd->priv->inhibitors, g_free);
+                close (systemd->priv->inhibit_fd);
+        }
+
         G_OBJECT_CLASS (gsm_systemd_parent_class)->finalize (object);
 }
 
@@ -130,9 +134,6 @@ gsm_systemd_init (GsmSystemd *manager)
                 g_object_unref (bus);
         }
 
-        manager->priv->authority = polkit_authority_get_sync (NULL, NULL);
-        manager->priv->subject = polkit_unix_session_new_for_process_sync (getpid (), NULL, NULL);
-
         sd_pid_get_session (getpid (), &manager->priv->session_id);
 
         res = g_dbus_proxy_call_sync (manager->priv->sd_proxy,
@@ -303,24 +304,24 @@ static gboolean
 gsm_systemd_can_restart (GsmSystem *system)
 {
         GsmSystemd *manager = GSM_SYSTEMD (system);
-        PolkitAuthorizationResult *res;
+        gchar *rv;
+        GVariant *res;
         gboolean can_restart;
 
-        res = polkit_authority_check_authorization_sync (manager->priv->authority,
-                                                         manager->priv->subject,
-                                                         "org.freedesktop.login1.reboot",
-                                                         NULL,
-                                                         POLKIT_CHECK_AUTHORIZATION_FLAGS_NONE,
-                                                         NULL,
-                                                         NULL);
-        if (res == NULL) {
-                return FALSE;
-        }
+        res = g_dbus_proxy_call_sync (manager->priv->sd_proxy,
+                                      "CanReboot",
+                                      NULL,
+                                      0,
+                                      G_MAXINT,
+                                      NULL,
+                                      NULL);
+        g_variant_get (res, "(s)", &rv);
+        g_variant_unref (res);
 
-        can_restart = polkit_authorization_result_get_is_authorized (res) ||
-                      polkit_authorization_result_get_is_challenge (res);
+        can_restart = g_strcmp0 (rv, "yes") == 0 ||
+                      g_strcmp0 (rv, "challenge") == 0;
 
-        g_object_unref (res);
+        g_free (rv);
 
         return can_restart;
 }
@@ -329,24 +330,24 @@ static gboolean
 gsm_systemd_can_stop (GsmSystem *system)
 {
         GsmSystemd *manager = GSM_SYSTEMD (system);
-        PolkitAuthorizationResult *res;
+        gchar *rv;
+        GVariant *res;
         gboolean can_stop;
 
-        res = polkit_authority_check_authorization_sync (manager->priv->authority,
-                                                         manager->priv->subject,
-                                                         "org.freedesktop.login1.power-off",
-                                                         NULL,
-                                                         POLKIT_CHECK_AUTHORIZATION_FLAGS_NONE,
-                                                         NULL,
-                                                         NULL);
-        if (res == NULL) {
-                return FALSE;
-        }
+        res = g_dbus_proxy_call_sync (manager->priv->sd_proxy,
+                                      "CanPowerOff",
+                                      NULL,
+                                      0,
+                                      G_MAXINT,
+                                      NULL,
+                                      NULL);
+        g_variant_get (res, "(s)", &rv);
+        g_variant_unref (res);
 
-        can_stop = polkit_authorization_result_get_is_authorized (res) ||
-                   polkit_authorization_result_get_is_challenge (res);
+        can_stop = g_strcmp0 (rv, "yes") == 0 ||
+                   g_strcmp0 (rv, "challenge") == 0;
 
-        g_object_unref (res);
+        g_free (rv);
 
         return can_stop;
 }
@@ -372,16 +373,221 @@ gsm_systemd_is_login_session (GsmSystem *system)
         return ret;
 }
 
+static gboolean
+gsm_systemd_can_suspend (GsmSystem *system)
+{
+        GsmSystemd *manager = GSM_SYSTEMD (system);
+        gchar *rv;
+        GVariant *res;
+        gboolean can_suspend;
+
+        res = g_dbus_proxy_call_sync (manager->priv->sd_proxy,
+                                      "CanSuspend",
+                                      NULL,
+                                      0,
+                                      G_MAXINT,
+                                      NULL,
+                                      NULL);
+        g_variant_get (res, "(s)", &rv);
+        g_variant_unref (res);
+
+        can_suspend = g_strcmp0 (rv, "yes") == 0 ||
+                      g_strcmp0 (rv, "challenge") == 0;
+
+        g_free (rv);
+
+        return can_suspend;
+}
+
+static gboolean
+gsm_systemd_can_hibernate (GsmSystem *system)
+{
+        GsmSystemd *manager = GSM_SYSTEMD (system);
+        gchar *rv;
+        GVariant *res;
+        gboolean can_hibernate;
+
+        res = g_dbus_proxy_call_sync (manager->priv->sd_proxy,
+                                      "CanHibernate",
+                                      NULL,
+                                      0,
+                                      G_MAXINT,
+                                      NULL,
+                                      NULL);
+        g_variant_get (res, "(s)", &rv);
+        g_variant_unref (res);
+
+        can_hibernate = g_strcmp0 (rv, "yes") == 0 ||
+                        g_strcmp0 (rv, "challenge") == 0;
+
+        g_free (rv);
+
+        return can_hibernate;
+}
+
+static void
+suspend_done (GObject      *source,
+              GAsyncResult *result,
+              gpointer      user_data)
+{
+        GDBusProxy *proxy = G_DBUS_PROXY (source);
+        GError *error = NULL;
+        GVariant *res;
+
+        res = g_dbus_proxy_call_finish (proxy, result, &error);
+
+        if (!res) {
+                g_warning ("Unable to suspend system: %s", error->message);
+                g_error_free (error);
+        } else {
+                g_variant_unref (res);
+        }
+}
+
+static void
+hibernate_done (GObject      *source,
+                GAsyncResult *result,
+              gpointer      user_data)
+{
+        GDBusProxy *proxy = G_DBUS_PROXY (source);
+        GError *error = NULL;
+        GVariant *res;
+
+        res = g_dbus_proxy_call_finish (proxy, result, &error);
+
+        if (!res) {
+                g_warning ("Unable to hibernate system: %s", error->message);
+                g_error_free (error);
+        } else {
+                g_variant_unref (res);
+        }
+}
+
+static void
+gsm_systemd_suspend (GsmSystem *system)
+{
+        GsmSystemd *manager = GSM_SYSTEMD (system);
+
+        g_dbus_proxy_call (manager->priv->sd_proxy,
+                           "Suspend",
+                           g_variant_new ("(b)", TRUE),
+                           0,
+                           G_MAXINT,
+                           NULL,
+                           hibernate_done,
+                           manager);
+}
+
+static void
+gsm_systemd_hibernate (GsmSystem *system)
+{
+        GsmSystemd *manager = GSM_SYSTEMD (system);
+
+        g_dbus_proxy_call (manager->priv->sd_proxy,
+                           "Hibernate",
+                           g_variant_new ("(b)", TRUE),
+                           0,
+                           G_MAXINT,
+                           NULL,
+                           suspend_done,
+                           manager);
+}
+
+static void
+inhibit_done (GObject      *source,
+              GAsyncResult *result,
+              gpointer      user_data)
+{
+        GDBusProxy *proxy = G_DBUS_PROXY (source);
+        GsmSystemd *manager = GSM_SYSTEMD (user_data);
+        GError *error = NULL;
+        GVariant *res;
+        GUnixFDList *fd_list = NULL;
+        gint idx;
+
+        res = g_dbus_proxy_call_with_unix_fd_list_finish (proxy, &fd_list, result, &error);
+
+        if (!res) {
+                g_warning ("Unable to inhibit system: %s", error->message);
+                g_error_free (error);
+        } else {
+                g_variant_get (res, "(h)", &idx);
+                manager->priv->inhibit_fd = g_unix_fd_list_get (fd_list, idx, &error);
+                if (manager->priv->inhibit_fd == -1) {
+                        g_warning ("Failed to receive system inhibitor fd: %s", error->message);
+                        g_error_free (error);
+                }
+                g_debug ("System inhibitor fd is %d", manager->priv->inhibit_fd);
+                g_object_unref (fd_list);
+                g_variant_unref (res);
+        }
+}
+
+static void
+gsm_systemd_add_inhibitor (GsmSystem        *system,
+                           const gchar      *id,
+                           GsmInhibitorFlag  flag)
+{
+        GsmSystemd *manager = GSM_SYSTEMD (system);
+
+        if ((flag & GSM_INHIBITOR_FLAG_SUSPEND) == 0)
+                return;
+
+        if (manager->priv->inhibitors == NULL) {
+                g_debug ("Adding system inhibitor");
+                g_dbus_proxy_call_with_unix_fd_list (manager->priv->sd_proxy,
+                                                     "Inhibit",
+                                                     g_variant_new ("(ssss)",
+                                                                    "sleep:shutdown",
+                                                                    g_get_user_name (),
+                                                                    "user session inhibited",
+                                                                    "block"),
+                                                     0,
+                                                     G_MAXINT,
+                                                     NULL,
+                                                     NULL,
+                                                     inhibit_done,
+                                                     manager);
+        }
+        manager->priv->inhibitors = g_slist_prepend (manager->priv->inhibitors, g_strdup (id));
+}
+
+static void
+gsm_systemd_remove_inhibitor (GsmSystem   *system,
+                              const gchar *id)
+{
+        GsmSystemd *manager = GSM_SYSTEMD (system);
+        GSList *l;
+
+        l = g_slist_find_custom (manager->priv->inhibitors, id, (GCompareFunc)g_strcmp0);
+        if (l == NULL)
+                return;
+
+        g_free (l->data);
+        manager->priv->inhibitors = g_slist_delete_link (manager->priv->inhibitors, l);
+        if (manager->priv->inhibitors == NULL) {
+                        g_debug ("Dropping system inhibitor");
+                        close (manager->priv->inhibit_fd);
+                        manager->priv->inhibit_fd = -1;
+        }
+}
+
 static void
 gsm_systemd_system_init (GsmSystemInterface *iface)
 {
         iface->can_switch_user = gsm_systemd_can_switch_user;
         iface->can_stop = gsm_systemd_can_stop;
         iface->can_restart = gsm_systemd_can_restart;
+        iface->can_suspend = gsm_systemd_can_suspend;
+        iface->can_hibernate = gsm_systemd_can_hibernate;
         iface->attempt_stop = gsm_systemd_attempt_stop;
         iface->attempt_restart = gsm_systemd_attempt_restart;
+        iface->suspend = gsm_systemd_suspend;
+        iface->hibernate = gsm_systemd_hibernate;
         iface->set_session_idle = gsm_systemd_set_session_idle;
         iface->is_login_session = gsm_systemd_is_login_session;
+        iface->add_inhibitor = gsm_systemd_add_inhibitor;
+        iface->remove_inhibitor = gsm_systemd_remove_inhibitor;
 }
 
 GsmSystemd *
